import collections
import os
import sys
import importlib
from glob import glob

import click
import toml

from spin import cmds as _cmds
from spin.cmds import util
from spin.sectioned_help import SectionedHelpGroup
from spin.color_format import ColorHelpFormatter


click.Context.formatter_class = ColorHelpFormatter


class DotDict(collections.UserDict):
    def __getitem__(self, key):
        subitem = self.data
        for subkey in key.split("."):
            try:
                subitem = subitem[subkey]
            except KeyError:
                raise KeyError(f"`{key}` not found in configuration") from None
        return subitem


def main():
    if not os.path.exists("pyproject.toml"):
        print("Error: cannot find [pyproject.toml]")
        sys.exit(1)

    with open("pyproject.toml") as f:
        try:
            toml_config = toml.load(f)
        except:
            print("Cannot parse [pyproject.toml]")
            sys.exit(1)

    project_config = toml_config.get("project", {})

    try:
        config = toml_config["tool"]["spin"]
    except KeyError:
        print("No configuration found in [pyproject.toml] for [tool.spin]")
        sys.exit(1)

    proj_name = project_config.get("name", config["package"])

    @click.group(help=f"Developer tool for {proj_name}", cls=SectionedHelpGroup)
    @click.pass_context
    def group(ctx):
        ctx.meta["config"] = DotDict(toml_config)
        ctx.meta["commands"] = ctx.command.section_commands
        ctx.show_default = True

    config_cmds = config["commands"]
    # Commands can be provided as a list, or as a dictionary
    # so that they can be sorted into sections
    if isinstance(config_cmds, list):
        config_cmds = {"Commands": config_cmds}

    # Backward compatibility workaround
    # Originally, you could specify any of these commands as `spin.cmd`
    # and we'd fetch it from util
    commands = {
        "spin.build": _cmds.meson.build,
        "spin.test": _cmds.meson.test,
        "spin.ipython": _cmds.meson.ipython,
        "spin.python": _cmds.meson.python,
        "spin.shell": _cmds.meson.shell,
    }

    for section, cmds in config_cmds.items():
        for cmd in cmds:
            if cmd not in commands:
                # First, see if we can directly import the command
                if not ":" in cmd:
                    path, func = cmd.rsplit(".", maxsplit=1)
                    try:
                        mod = importlib.import_module(path)
                    except ImportError:
                        print(
                            f"!! Could not import module `{path}` to load command `{cmd}`"
                        )
                        continue
                else:
                    try:
                        path, func = cmd.split(":")
                        spec = importlib.util.spec_from_file_location(
                            "custom_mod", path
                        )
                        mod = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(mod)
                    except FileNotFoundError:
                        print(
                            f"!! Could not find file `{path}` to load custom command `{cmd}`.\n"
                        )
                        continue
                    except Exception as e:
                        print(
                            f"!! Could not import file `{path}` to load custom command `{cmd}`.\n"
                        )
                        raise e

                try:
                    cmd_func = getattr(mod, func)
                except AttributeError:
                    print(f"!! Could not load command `{func}` from file `{path}`.\n")
                    continue

                commands[cmd] = cmd_func

            group.add_command(commands[cmd], section=section)

    try:
        group()
    except Exception as e:
        print(f"{e}; aborting.")
        sys.exit(1)


if __name__ == "__main__":
    main()



import os
import sys
import subprocess
import shlex
import sysconfig
from pathlib import Path

import click


def run(
    cmd, cwd=None, replace=False, sys_exit=True, output=True, echo=True, *args, **kwargs
):
    """Run a shell command.

    Parameters
    ----------
    cwd : str
        Change to this directory before execution.
    replace : bool
        Whether to replace the current process.
    sys_exit : bool
        Whether to exit if the shell command returns with error != 0.
    output : bool
        Whether to display output as the process runs.
        If set to ``False``, can be accessed afterwards as
        ``p.stdout``.
        If `sys_exit` is True and the process fails, output is printed
        regardless.
    echo : bool
        Whether or not to echo commands.

    Other arguments and keywords are passed directly to `subprocess.run`.

    """
    if cwd:
        if echo:
            click.secho(f"$ cd {cwd}", bold=True, fg="bright_blue")
        os.chdir(cwd)

    cmdstr = " ".join(shlex.quote(arg) for arg in cmd)
    if echo:
        click.secho(f"$ {cmdstr}", bold=True, fg="bright_blue")

    if output is False:
        output_kwargs = {"stdout": subprocess.PIPE, "stderr": subprocess.STDOUT}
        kwargs = {**output_kwargs, **kwargs}

    if replace:
        os.execvp(cmd[0], cmd)
        print(f"Failed to launch `{cmd}`")
        sys.exit(-1)
    else:
        p = subprocess.run(cmd, *args, **kwargs)
        if p.returncode != 0 and sys_exit:
            # Output was suppressed, but the process failed, so print it anyway
            if output is False:
                print(p.stdout.decode("utf-8"), end="")
            sys.exit(p.returncode)
        return p


def get_config():
    """Return the pyproject.toml as a dictionary."""
    return click.get_current_context().meta["config"]


def get_commands():
    """Return a list of all commands.

    Returns
    -------
    cmds : dict
       The keys of the dictionary are the section names, the values
       lists of commands.

       If no sections are provided, all commands are listed under the
       ``commands`` key.
    """
    return click.get_current_context().meta["commands"]



import re
import shutil

import click


class RegexpFormatter:
    def __init__(self, rules, default=None):
        if default is None:
            default = {}
        self.rules = {re.compile(expr): format for (expr, format) in rules.items()}
        self.default_style = click.style("", **default, reset=False)
        self.reset_style = click.style("", reset=True)

    def __call__(self, txt):
        for expr, format in self.rules.items():
            txt = expr.sub(click.style("\\g<0>", **format) + self.default_style, txt)
        txt = self.default_style + txt + self.reset_style
        return txt


class ColorHelpFormatter(click.formatting.HelpFormatter):
    def __init__(self, *args, **kwargs):
        kwargs["width"] = shutil.get_terminal_size()[0]
        super().__init__(*args, **kwargs)

    def write_heading(self, heading):
        super().write_heading(click.style(heading, dim=True))

    def write_dl(self, items):
        """Print definition list"""
        key_fmt = RegexpFormatter(
            {
                " [A-Z]+": {"bold": True, "fg": "yellow"},
                r"\-[a-z]{1}(?=[ ,]{1})": {"fg": "green"},
            },
            default={"bold": True, "fg": "cyan"},
        )
        val_fmt = RegexpFormatter({r"\[default: .*?\]": {"dim": True}})
        items = [(key_fmt(key), val_fmt(val)) for (key, val) in items]
        super().write_dl(items)

    def write_usage(self, prog, args="", prefix=None):
        fmt = RegexpFormatter(
            {r"(?<=\[)[A-Z_]+(?=\])": {"fg": "cyan"}}, default={"bold": True}
        )
        super().write_usage(
            click.style(prog, bold=True, fg="white"),
            click.style(fmt(args), bold=True),
            prefix=click.style("Usage: ", bold=True, fg="yellow"),
        )



import json
import click
from spin import util


@click.command()
@click.option("-f", "--flag")
def example(flag):
    """🧪 Example custom command.

    Accepts arbitrary flags, and shows how to access `pyproject.toml`
    config.
    """
    click.secho("Running example custom command", bold=True, fg="bright_blue")
    print()
    config = util.get_config()
    commands = util.get_commands()
    click.secho("Flag provided with --flag is: ", fg="yellow", nl=False)
    print(flag or None)

    click.secho("\nTool config is:", fg="yellow")
    print(json.dumps(config["tool.spin"], indent=2))
